from tests.test_hyper_utils import get_pyarrow_table
from tableauhyperapi import HyperProcess, Telemetry, Connection, CreateMode
from packages.hyper_file import HyperFile
import pyarrow as pa
import datetime as dt
import pytest
import os

FILENAME = "test.hyper"
QRY = 'SELECT COUNT(*) FROM "Extract"."Extract"'


@pytest.fixture
def create_hyper_file(get_pyarrow_table):
    def _method(hyper_filename):
        PARQUET = "parquet"
        df = get_pyarrow_table
        filename = str(dt.datetime.today().strftime("%Y-%m-%d")) + "." + PARQUET
        pa.parquet.write_table(df, filename)
        hf = HyperFile("", PARQUET)
        hf.create_hyper_file(hyper_filename)
        return hf

    return _method


def test_create_hyper_file(create_hyper_file):
    create_hyper_file(FILENAME)
    with HyperProcess(Telemetry.SEND_USAGE_DATA_TO_TABLEAU) as hyper:
        with Connection(hyper.endpoint, FILENAME, CreateMode.NONE) as con:
            rows = con.execute_scalar_query(QRY)
    os.remove(FILENAME)
    assert rows == 2


def test_delete_rows(create_hyper_file):
    COLUMN_NAME = "date32"
    hf = create_hyper_file(FILENAME)
    count = hf.delete_rows(FILENAME, COLUMN_NAME, 1)
    os.remove(FILENAME)
    assert count == 1


def test_append_rows(create_hyper_file):
    hf = create_hyper_file(FILENAME)
    hf.append_rows(FILENAME)
    with HyperProcess(Telemetry.SEND_USAGE_DATA_TO_TABLEAU) as hyper:
        with Connection(hyper.endpoint, FILENAME, CreateMode.NONE) as con:
            rows = con.execute_scalar_query(QRY)
    os.remove(FILENAME)
    assert rows == 4



from tableauhyperapi import SqlType, TableDefinition, NULLABLE, TableName, HyperProcess, Telemetry, \
    Connection, CreateMode, escape_name
from pathlib import Path
from packages.time_decorator import timeit
from pyarrow import ChunkedArray
from pyarrow.parquet import ParquetFile
import pyarrow as pa
import os
import subprocess
import logging
import argparse

class GCSFuse:

    def __init__(self, bucket_name : str, mnt_dir : str) -> None:
        """

        Args:
            bucket_name (str): bucket name
            mnt_dir (str): path where the bucket will be mounted
        """
        self.bucket_name = bucket_name
        self.mnt_dir = mnt_dir

    def mount(self) -> None:
        """Mount GCS bucket as a path on the machine"""
        os.makedirs(self.mnt_dir, exist_ok=True)
        exec = subprocess.run(['gcsfuse', '--implicit-dirs', self.bucket_name, self.mnt_dir],
                stdout=subprocess.DEVNULL)
        if exec.returncode != 0:
            raise ValueError(f'The mount command returned error {exec.returncode}. Check the whether the \
                bucket {self.bucket_name} or the mount dir {self.mnt_dir} is correct.')
        return exec.returncode


class HyperSuite:

    def __init__(self, bucket_name : str, parquet_dir : str, hyper_db_path : str) -> None:
        self.bucket_name = bucket_name
        self.parquet_dir = parquet_dir
        self.hyper_db_path = hyper_db_path
        self.g_fuse = GCSFuse(bucket_name, './gcs/')
        self.filenames = self._get_filenames()

    def _get_filenames(self) -> list[str]:
        gcs_path = f'gs://{self.bucket_name}/{self.parquet_dir}/'
        process = subprocess.run(['gsutil', 'ls', gcs_path], shell=True, capture_output=True, text=True)
        filenames = [os.path.basename(path) for path in process.stdout.splitlines()]
        if len(filenames) == 0:
            raise ValueError(f'Error! The bucket path {gcs_path} returned no files!')
        return filenames

    def copy_parquet_to_hyper_file(self) -> None:
        """Function that copies data from a Parquet file to a .hyper file."""
        if os.path.exists(self.hyper_db_path):
            os.remove(self.hyper_db_path)
        with HyperProcess(telemetry=Telemetry.DO_NOT_SEND_USAGE_DATA_TO_TABLEAU) as hp:
            with Connection(endpoint=hp.endpoint,
                            database=Path(self.hyper_db_path + '.hyper'),
                            create_mode=CreateMode.CREATE_AND_REPLACE) as connection:
                table_definition = self._get_table_def(ParquetFile(self.filenames[0]))
                connection.catalog.create_schema(schema=table_definition.table_name.schema_name)
                connection.catalog.create_table(table_definition=table_definition)
                total_rows = 0
                for parquet_path in self.filenames:
                    try:
                        self.g_fuse.mount()
                        copy_command = f"COPY \"Extract\".\"Extract\" from '{parquet_path}' with (format parquet)"
                        count = connection.execute_command(copy_command)
                        total_rows += count
                    except Exception:
                        logging.warning(f'File {os.path.basename(parquet_path)} is empty.')
                logging.info(f'Process completed with {total_rows} rows added.')
                return total_rows

    def delete_rows_from_hyper_file(self, days_to_delete : int, date_column : str) -> None:
        """Function that copies data from a Parquet file to a .hyper file.

        Args:
            days_to_delete (int): the window in days to be deleted from the database
            date_column (str): name of date column to be used in the incremental strategy
        """
        with HyperProcess(telemetry=Telemetry.DO_NOT_SEND_USAGE_DATA_TO_TABLEAU) as hp:
            with Connection(endpoint=hp.endpoint,
                            database=Path(self.hyper_db_path),
                            create_mode=CreateMode.NONE) as connection:
                self.g_fuse.mount()
                delete_command = f'DELETE FROM \"Extract\".\"Extract\" WHERE {escape_name(date_column)} >= CURRENT_DATE - {days_to_delete}'
                count = connection.execute_command(delete_command)
                logging.info(f'Process completed with {count} rows deleted.')

    def insert_hyper_records(self) -> None:
        """Function to insert records to an existing hyper file"""
        with HyperProcess(telemetry=Telemetry.DO_NOT_SEND_USAGE_DATA_TO_TABLEAU) as hp:
            with Connection(endpoint=hp.endpoint,
                            database=Path(self.hyper_db_path),
                            create_mode=CreateMode.NONE) as connection:
                total_rows = 0
                for parquet_path in self.parquet_dir:
                    try:
                        self.g_fuse.mount()
                        copy_command = f"COPY \"Extract\".\"Extract\" from '{parquet_path}' with (format parquet)"
                        count = connection.execute_command(copy_command)
                        total_rows += count
                    except Exception as e:
                        logging.warning(f'File {os.path.basename(parquet_path)} is empty.')
                logging.info(f'Process completed with {total_rows} rows added.')

@timeit
def main() -> None:
    logging.basicConfig(level=logging.INFO)
    logging.info('Starting create single hyper process.')

    CLI = argparse.ArgumentParser()
    CLI.add_argument('--parquet_folder_path', type=str)
    CLI.add_argument('--bucket_name', type=str)
    CLI.add_argument('--hyper_file_path', type=str)
    CLI.add_argument('--days_to_delete', type=int, default=1)
    CLI.add_argument('--date_column', type=str, default='')
    CLI.add_argument('--func', type=str, default='incremental')
    args = CLI.parse_args()

    parquet_folder  = args.parquet_folder_path   #'PARQUET_1'
    bucket_name     = args.bucket_name          #'br-apps-dolphin-ddp-dev-comercial'
    hyper_file_path = args.hyper_file_path      #'/mnt/daily_sales.hyper'
    days_to_delete  = args.days_to_delete       #30
    date_column     = args.date_column          # DATE_KEY
    func            = args.func                 # full/incremental

    hs = HyperSuite(bucket_name, parquet_folder, hyper_file_path)
    if func == 'full':
        hs.copy_parquet_to_hyper_file()
    elif func == 'incremental':
        hs.delete_rows_from_hyper_file(days_to_delete, date_column)
        hs.insert_hyper_records()

if __name__ == '__main__':
    gs = GCSFuse('hyper-suite-cr4', './gcs')
    print(gs.mount())


from tableauhyperapi import HyperProcess, Telemetry, Connection, CreateMode, escape_name
from pyarrow.parquet import ParquetFile
import os
import logging as log
import packages.hyper_utils as hu


class HyperFile:
    def __init__(self, parquet_folder: str, file_extension: str = None) -> None:
        """Requires parquet folder and parquet file extension if any

        Args:
            parquet_folder (str): path to folder with parquet files
            file_extension (str, optional): parquet file extension
                without the dot. Defaults to None.
        """
        self.parquet_folder = parquet_folder
        self.file_extension = file_extension

    def create_hyper_file(self, hyper_path: str) -> int:
        """Create hyper file based on files within parquet_folder

        Args:
            hyper_path (str): hyper destination with file name.
                Eg: /path/file.hyper

        Returns:
            int: number of affected rows
        """
        if os.path.exists(hyper_path):
            os.remove(hyper_path)
        files = hu.get_parquet_files(self.parquet_folder, self.file_extension)
        telemetry = Telemetry.DO_NOT_SEND_USAGE_DATA_TO_TABLEAU
        create_mode = CreateMode.CREATE_AND_REPLACE
        with HyperProcess(telemetry=telemetry) as hp:
            with Connection(
                endpoint=hp.endpoint, database=hyper_path, create_mode=create_mode
            ) as conn:
                table_definition = hu.get_table_def(ParquetFile(files[0]))
                schema = table_definition.table_name.schema_name
                conn.catalog.create_schema(schema=schema)
                conn.catalog.create_table(table_definition=table_definition)
                total_rows = 0
                for file in files:
                    try:
                        copy_command = f'COPY "Extract"."Extract" from \'{file}\' with (format parquet)'
                        count = conn.execute_command(copy_command)
                        total_rows += count
                    except Exception as e:
                        log.warning(
                            f"File {os.path.basename(file)} could not be processed. {e}"
                        )
                        log.info(f"Error message: {e}")
                log.info(f"Process completed with {total_rows} rows added.")
                return total_rows

    def delete_rows(
        self, hyper_path: str, date_column: str, days_to_delete: int
    ) -> int:
        """Delete rows from a hyper based on days before a date to delete.

        Args:
            hyper_path (str): hyper file path. Eg: path/hyper.file
            days_to_delete (int): the window in days to delete in the database
            date_column (str): date column to use in the incremental strategy
        Returns:
            int: number of deleted rows
        """
        telemetry = Telemetry.DO_NOT_SEND_USAGE_DATA_TO_TABLEAU
        with HyperProcess(telemetry=telemetry) as hp:
            with Connection(
                endpoint=hp.endpoint, database=hyper_path, create_mode=CreateMode.NONE
            ) as connection:
                qry = f'DELETE FROM "Extract"."Extract" WHERE {escape_name(date_column)} >= CURRENT_DATE - {days_to_delete}'  # noqa
                count = connection.execute_command(qry)
                log.info(f"Process completed with {count} rows deleted.")
        return count

    def append_rows(self, hyper_path: str) -> int:
        """Append rows from parquet files into an existing hyper file

        Args:
            hyper_path (str): hyper file path. Eg: path/hyper.file

        Returns:
            int: number of appended rows
        """
        telemetry = Telemetry.DO_NOT_SEND_USAGE_DATA_TO_TABLEAU
        with HyperProcess(telemetry=telemetry) as hp:
            with Connection(
                endpoint=hp.endpoint, database=hyper_path, create_mode=CreateMode.NONE
            ) as connection:
                total_rows = 0
                files = hu.get_parquet_files(self.parquet_folder, self.file_extension)
                for parquet_path in files:
                    try:
                        copy_command = f'COPY "Extract"."Extract" from \'{parquet_path}\' with (format parquet)'
                        count = connection.execute_command(copy_command)
                        total_rows += count
                    except Exception as e:
                        log.warning(
                            f"File {os.path.basename(parquet_path)} could not be processed. {e}"
                        )
                        log.info(f"Error message: {e}")
                log.info(f"Process completed with {total_rows} rows added.")
        return total_rows



import packages.hyper_utils as hu
import pyarrow as pa
import pytest
import datetime as dt
import os
from tableauhyperapi import SqlType

STRF_TIME = "%Y-%m-%d"


@pytest.fixture
def get_pyarrow_table():
    A = "a"
    B = "b"
    US = "us"
    UTF8 = "utf-8"
    COL_NAMES = [
        "int8",
        "int16",
        "int32",
        "int64",
        "string",
        "float32",
        "float64",
        "bool",
        "timestamp",
        "date32",
        "date64",
        "binary",
        "decimal128",
    ]
    array = [
        pa.array([1, 2], type=pa.int8()),
        pa.array([1, 2], type=pa.int16()),
        pa.array([1, 2], type=pa.int32()),
        pa.array([1, 2], type=pa.int64()),
        pa.array([A, B], type=pa.string()),
        pa.array([1.0, 1.5], type=pa.float32()),
        pa.array([1.0, 1.5], type=pa.float64()),
        pa.array([True, False], type=pa.bool_()),
        pa.array(
            [dt.datetime(2023, 1, 1, 0, 0, 0), dt.datetime.now()], type=pa.timestamp(US)
        ),
        pa.array([dt.date(2023, 1, 1), dt.date.today()], type=pa.date32()),
        pa.array([dt.date(2023, 1, 1), dt.date.today()], type=pa.date64()),
        pa.array([A.encode(UTF8), B.encode(UTF8)], type=pa.binary()),
        pa.array([1234, 1234], type=pa.decimal128(7, 3)),
    ]
    yield pa.table(array, names=COL_NAMES)


@pytest.fixture
def get_pyarrow_schema(get_pyarrow_table):
    yield get_pyarrow_table.schema


def test_convert_struct_field(get_pyarrow_schema):
    gps = get_pyarrow_schema
    assert hu._convert_struct_field(gps[0]).type == SqlType.small_int()
    assert hu._convert_struct_field(gps[1]).type == SqlType.small_int()
    assert hu._convert_struct_field(gps[2]).type == SqlType.int()
    assert hu._convert_struct_field(gps[3]).type == SqlType.big_int()
    assert hu._convert_struct_field(gps[4]).type == SqlType.text()
    assert hu._convert_struct_field(gps[5]).type == SqlType.double()
    assert hu._convert_struct_field(gps[6]).type == SqlType.double()
    assert hu._convert_struct_field(gps[7]).type == SqlType.bool()
    assert hu._convert_struct_field(gps[8]).type == SqlType.timestamp()
    assert hu._convert_struct_field(gps[9]).type == SqlType.date()
    assert hu._convert_struct_field(gps[10]).type == SqlType.date()
    assert hu._convert_struct_field(gps[11]).type == SqlType.bytes()
    assert hu._convert_struct_field(gps[12]).type == SqlType.numeric(7, 3)


def test_get_table_def(get_pyarrow_table):
    df = get_pyarrow_table
    now = str(dt.datetime.today().strftime(STRF_TIME))
    pa.parquet.write_table(df, now)
    with pa.parquet.ParquetFile(now) as file:
        table_def = hu.get_table_def(file)
    os.remove(now)
    assert table_def.columns[0].type == SqlType.small_int()
    assert table_def.columns[1].type == SqlType.small_int()
    assert table_def.columns[2].type == SqlType.int()
    assert table_def.columns[3].type == SqlType.big_int()
    assert table_def.columns[4].type == SqlType.text()
    assert table_def.columns[5].type == SqlType.double()
    assert table_def.columns[6].type == SqlType.double()
    assert table_def.columns[7].type == SqlType.bool()
    assert table_def.columns[8].type == SqlType.timestamp()
    assert table_def.columns[9].type == SqlType.date()
    assert table_def.columns[10].type == SqlType.date()
    assert table_def.columns[11].type == SqlType.bytes()
    assert table_def.columns[12].type == SqlType.numeric(7, 3)


def test_get_parquet_files(get_pyarrow_table):
    PARQUET = "parquet"
    df = get_pyarrow_table
    now = str(dt.datetime.today().strftime(STRF_TIME))
    extension = "." + PARQUET
    filename = now + extension
    pa.parquet.write_table(df, filename)
    files = hu.get_parquet_files("", extension.replace(".", ""))
    os.remove(filename)
    assert len(files) == 1



from pyarrow import ChunkedArray
from tableauhyperapi import SqlType, TableDefinition, NULLABLE, NOT_NULLABLE, TableName
from pyarrow.parquet import ParquetFile
import glob
import pyarrow as pa


def _convert_struct_field(column: ChunkedArray) -> TableDefinition.Column:
    """Converts a Pyarrow Column type to a Tableau Hyper SqlType

    Args:
        column (ChunkedArray): Pyarrow column

    Raises:
        ValueError: raises error if the column type is not found.

    Returns:
        Column: Column with Hyper SqlType
    """
    S = "s"
    MS = "ms"
    NS = "ns"
    US = "us"
    DECIMAL = "decimal"
    if column.type == pa.string():
        sql_type = SqlType.text()
    elif column.type in [pa.date32(), pa.date64()]:
        sql_type = SqlType.date()
    elif column.type in [pa.float32(), pa.float64()]:
        sql_type = SqlType.double()
    elif column.type in [pa.int8(), pa.int16()]:
        sql_type = SqlType.small_int()
    elif column.type == pa.int32():
        sql_type = SqlType.int()
    elif column.type == pa.int64():
        sql_type = SqlType.big_int()
    elif column.type == pa.bool_():
        sql_type = SqlType.bool()
    elif column.type in [
        pa.timestamp(S),
        pa.timestamp(MS),
        pa.timestamp(US),
        pa.timestamp(NS),
    ]:
        sql_type = SqlType.timestamp()
    elif column.type == pa.binary():
        sql_type = SqlType.bytes()
    elif str(column.type).startswith(DECIMAL):
        precision = column.type.precision
        scale = column.type.scale
        sql_type = SqlType.numeric(precision, scale)
    else:
        raise ValueError(
            f"Invalid StructField datatype for column `{column.name}` : {column.type}"
        )
    nullable = NULLABLE if column.nullable else NOT_NULLABLE
    return TableDefinition.Column(name=column.name, type=sql_type, nullability=nullable)


def get_table_def(df: ParquetFile) -> TableDefinition:
    """Returns a Tableau TableDefintion given a Parquet file

    Args:
        df (ParquetFile): Pyarrow dataframe

    Returns:
        TableDefinition: Hyper Table Definition with schema
    """
    schema = df.schema_arrow
    cols = list(map(_convert_struct_field, schema))
    return TableDefinition(table_name=TableName("Extract", "Extract"), columns=cols)


def get_parquet_files(parquet_folder: str, parquet_extension: str = None) -> list[str]:
    """Get list of parquet files in a folder

    Args:
        parquet_folder (str): path where the parquet files are
        parquet_extension (str, optional): If the parquet files has some
            extension in its files. Eg.: parquet. Defaults to None.

    Raises:
        ValueError: raises error if the folder has no parquet files

    Returns:
        list[str]: list of filenames
    """
    ext = f"*.{parquet_extension}" if parquet_extension is not None else "*"
    files = glob.glob(parquet_folder + ext)
    if len(files) == 0:
        raise ValueError(
            f"Error! The parquet_folder: {parquet_folder} returned no files!"
        )
    return files



import tableauserverclient as TSC
import logging


class TableauServerUtils:
    def __init__(self, server_address: str, token_name: str, token_value: str) -> None:
        """TableauServerUtils constructor

        Args:
            server_address (str): tableau server address
            token_name (str): token name
            token_value (str): token value
        """
        self.server = TSC.Server(server_address, use_server_version=True)
        self.tableau_auth = TSC.PersonalAccessTokenAuth(token_name, token_value)

    def get_project_id(self, project_name: str) -> str:
        """Get project id by name

        Args:
            project_name (str): project name

        Raises:
            ValueError: raises error whether the functions find more than 1
             project with the same project_name

        Returns:
            str: project id
        """
        logging.info(f"Signing into the server {self.server.baseurl}")
        with self.server.auth.sign_in(self.tableau_auth):
            req_options = TSC.RequestOptions()
            req_options.filter.add(
                TSC.Filter(
                    TSC.RequestOptions.Field.Name,
                    TSC.RequestOptions.Operator.Equals,
                    project_name,
                )
            )
            projects = list(TSC.Pager(self.server.projects, req_options))
            if len(projects) > 1:
                raise ValueError("The project name is not unique.")
        return projects[0].id

    def publish_hyper(
        self, project_id: str, hyper_path: str, mode: str = "overwrite"
    ) -> None:
        """Publish hyper file into the tableau server

        Args:
            project_id (str): project id
            hyper_path (str): hyper file path. Eg: path/hyper.file
            mode (str): publish mode. Accept overwrite or append mode.
            Defaults to overwrite.
        """
        OVERWRITE = "overwrite"
        APPEND = "append"
        logging.info(f"Signing into the server {self.server.baseurl}")
        with self.server.auth.sign_in(self.tableau_auth):
            if mode == OVERWRITE:
                publish_mode = TSC.Server.PublishMode.Overwrite
            elif mode == APPEND:
                publish_mode = TSC.Server.PublishMode.Append
            else:
                raise ValueError(f"Mode must be overwrite or append. Received {mode}")
            datasource = TSC.DatasourceItem(project_id=project_id)
            logging.info("Publishing Hyper file into the server!")
            ds = self.server.datasources.publish(datasource, hyper_path, publish_mode)
            logging.info(f"Datasource published on ID: {ds.id}")
        logging.info("Job Finished.")



from functools import wraps
import logging
import time
from typing import Callable


def timeit(func: Callable) -> Callable:
    """Decorator to measure time of execution of a function

    Args:
        func (Callable): function to be measured

    Returns:
        Callable: measured function
    """

    @wraps(func)
    def timeit_wrapper(*args, **kwargs) -> Callable:
        start_time = time.perf_counter()
        result = func(*args, **kwargs)
        end_time = time.perf_counter()
        total_time = end_time - start_time
        logging.info(f"Function {func.__name__} took {total_time:.2f} seconds")
        return result

    return timeit_wrapper
